@@ -11,7 +11,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclWinFile.c,v 1.39 2002/07/20 01:01:41 vincentdarley Exp $
+ * RCS: @(#) $Id: tclWinFile.c,v 1.44.2.6 2003/10/03 17:45:37 vincentdarley Exp $
  */
 
 //#define _WIN32_WINNT  0x0500
@@ -128,6 +128,26 @@ typedef struct {
     WCHAR  dummyBuf[MAX_PATH*3];
 } DUMMY_REPARSE_BUFFER;
 
+#if defined(_MSC_VER) && ( _MSC_VER <= 1100 )
+#define HAVE_NO_FINDEX_ENUMS
+#elif !defined(_WIN32_WINNT) || (_WIN32_WINNT < 0x0400)
+#define HAVE_NO_FINDEX_ENUMS
+#endif
+
+#ifdef HAVE_NO_FINDEX_ENUMS
+/* These two aren't in VC++ 5.2 headers */
+typedef enum _FINDEX_INFO_LEVELS {
+	FindExInfoStandard,
+	FindExInfoMaxInfoLevel
+} FINDEX_INFO_LEVELS;
+typedef enum _FINDEX_SEARCH_OPS {
+	FindExSearchNameMatch,
+	FindExSearchLimitToDirectories,
+	FindExSearchLimitToDevices,
+	FindExSearchMaxSearchOp
+} FINDEX_SEARCH_OPS;
+#endif /* HAVE_NO_FINDEX_ENUMS */
+
 /* Other typedefs required by this code */
 
 static time_t		ToCTime(FILETIME fileTime);
@@ -141,19 +161,22 @@ typedef NET_API_STATUS NET_API_FUNCTION NETAPIBUFFERFREEPROC
 typedef NET_API_STATUS NET_API_FUNCTION NETGETDCNAMEPROC
 	(LPWSTR servername, LPWSTR domainname, LPBYTE *bufptr);
 
+extern Tcl_FSDupInternalRepProc NativeDupInternalRep;
+
 /*
  * Declarations for local procedures defined in this file:
  */
 
 static int NativeAccess(CONST TCHAR *path, int mode);
 static int NativeStat(CONST TCHAR *path, Tcl_StatBuf *statPtr, int checkLinks);
+static unsigned short NativeStatMode(DWORD attr, int checkLinks, int isExec);
 static int NativeIsExec(CONST TCHAR *path);
 static int NativeReadReparse(CONST TCHAR* LinkDirectory,
 			     REPARSE_DATA_BUFFER* buffer);
 static int NativeWriteReparse(CONST TCHAR* LinkDirectory,
 			      REPARSE_DATA_BUFFER* buffer);
-static int NativeMatchType(CONST char *name, int nameLen,
-			   CONST TCHAR* nativeName, Tcl_GlobTypeData *types);
+static int NativeMatchType(int isDrive, DWORD attr, CONST TCHAR* nativeName,
+			   Tcl_GlobTypeData *types);
 static int WinIsDrive(CONST char *name, int nameLen);
 static Tcl_Obj* WinReadLink(CONST TCHAR* LinkSource);
 static Tcl_Obj* WinReadLinkDirectory(CONST TCHAR* LinkDirectory);
@@ -162,7 +185,6 @@ static int WinLink(CONST TCHAR* LinkSource, CONST TCHAR* LinkTarget,
 static int WinSymLinkDirectory(CONST TCHAR* LinkDirectory,
 			       CONST TCHAR* LinkTarget);
 
-
 /*
  *--------------------------------------------------------------------
  *
@@ -248,8 +270,7 @@ WinLink(LinkSource, LinkTarget, linkAction)
  *
  * WinReadLink
  *
- * What does 'LinkSource' point to?  We need the original 'pathPtr'
- * just so we can construct a path object in the correct filesystem.
+ * What does 'LinkSource' point to?
  *--------------------------------------------------------------------
  */
 static Tcl_Obj*
@@ -393,7 +414,7 @@ TclWinSymLinkDelete(LinkOriginal, linkOnly)
     DUMMY_REPARSE_BUFFER dummy;
     REPARSE_DATA_BUFFER *reparseBuffer = (REPARSE_DATA_BUFFER*)&dummy;
     HANDLE hFile;
-    int returnedLength;
+    DWORD returnedLength;
     memset(reparseBuffer, 0, sizeof(DUMMY_REPARSE_BUFFER));
     reparseBuffer->ReparseTag = IO_REPARSE_TAG_MOUNT_POINT;
     hFile = (*tclWinProcs->createFileProc)(LinkOriginal, GENERIC_WRITE, 0,
@@ -428,7 +449,11 @@ TclWinSymLinkDelete(LinkOriginal, linkOnly)
  *
  * Assumption that LinkDirectory is a valid, existing directory.
  *
- * Returns a Tcl_Obj with refCount of 1 (i.e. owned by the caller).
+ * Returns a Tcl_Obj with refCount of 1 (i.e. owned by the caller),
+ * or NULL if anything went wrong.
+ *
+ * In the future we should enhance this to return a path object
+ * rather than a string.
  *--------------------------------------------------------------------
  */
 static Tcl_Obj*
@@ -456,28 +481,77 @@ WinReadLinkDirectory(LinkDirectory)
 	    Tcl_DString ds;
 	    CONST char *copy;
 	    int len;
+	    int offset = 0;
+
+	    /*
+	     * Certain native path representations on Windows have a
+	     * special prefix to indicate that they are to be treated
+	     * specially.  For example extremely long paths, or symlinks,
+	     * or volumes mounted inside directories.
+	     *
+	     * There is an assumption in this code that 'wide' interfaces
+	     * are being used (see tclWin32Dll.c), which is true for the
+	     * only systems which support reparse tags at present.  If
+	     * that changes in the future, this code will have to be
+	     * generalised.
+	     */
+	    if (reparseBuffer->SymbolicLinkReparseBuffer.PathBuffer[0]
+		                                                 == L'\\') {
+		/* Check whether this is a mounted volume */
+		if (wcsncmp(reparseBuffer->SymbolicLinkReparseBuffer.PathBuffer,
+			    L"\\??\\Volume{",11) == 0) {
+		    char drive;
+		    /*
+		     * There is some confusion between \??\ and \\?\ which
+		     * we have to fix here.  It doesn't seem very well
+		     * documented.
+		     */
+		    reparseBuffer->SymbolicLinkReparseBuffer
+		                                      .PathBuffer[1] = L'\\';
+		    /*
+		     * Check if a corresponding drive letter exists, and
+		     * use that if it is found
+		     */
+		    drive = TclWinDriveLetterForVolMountPoint(reparseBuffer
+					->SymbolicLinkReparseBuffer.PathBuffer);
+		    if (drive != -1) {
+			char driveSpec[3] = {
+			    drive, ':', '\0'
+			};
+			retVal = Tcl_NewStringObj(driveSpec,2);
+			Tcl_IncrRefCount(retVal);
+			return retVal;
+		    }
+		    /*
+		     * This is actually a mounted drive, which doesn't
+		     * exists as a DOS drive letter.  This means the path
+		     * isn't actually a link, although we partially treat
+		     * it like one ('file type' will return 'link'), but
+		     * then the link will actually just be treated like
+		     * an ordinary directory.  I don't believe any
+		     * serious inconsistency will arise from this, but it
+		     * is something to be aware of.
+		     */
+		    Tcl_SetErrno(EINVAL);
+		    return NULL;
+		} else if (wcsncmp(reparseBuffer->SymbolicLinkReparseBuffer
+				   .PathBuffer, L"\\\\?\\",4) == 0) {
+		    /* Strip off the prefix */
+		    offset = 4;
+		} else if (wcsncmp(reparseBuffer->SymbolicLinkReparseBuffer
+				   .PathBuffer, L"\\??\\",4) == 0) {
+		    /* Strip off the prefix */
+		    offset = 4;
+		}
+	    }
 
 	    Tcl_WinTCharToUtf(
 		(CONST char*)reparseBuffer->SymbolicLinkReparseBuffer.PathBuffer,
-		(int)reparseBuffer->SymbolicLinkReparseBuffer.SubstituteNameLength,
-		&ds);
+		(int)reparseBuffer->SymbolicLinkReparseBuffer
+		.SubstituteNameLength, &ds);
 
-	    copy = Tcl_DStringValue(&ds);
-	    len = Tcl_DStringLength(&ds);
-	    /*
-	     * Certain native path representations on Windows have this special
-	     * prefix to indicate that they are to be treated specially.  For
-	     * example extremely long paths, or symlinks
-	     */
-	    if (*copy == '\\') {
-		if (0 == strncmp(copy,"\\??\\",4)) {
-		    copy += 4;
-		    len -= 4;
-		} else if (0 == strncmp(copy,"\\\\?\\",4)) {
-		    copy += 4;
-		    len -= 4;
-		}
-	    }
+	    copy = Tcl_DStringValue(&ds)+offset;
+	    len = Tcl_DStringLength(&ds)-offset;
 	    retVal = Tcl_NewStringObj(copy,len);
 	    Tcl_IncrRefCount(retVal);
 	    Tcl_DStringFree(&ds);
@@ -506,7 +580,7 @@ NativeReadReparse(LinkDirectory, buffer)
     REPARSE_DATA_BUFFER* buffer;  /* Pointer to buffer. Cannot be NULL */
 {
     HANDLE hFile;
-    int returnedLength;
+    DWORD returnedLength;
 
     hFile = (*tclWinProcs->createFileProc)(LinkDirectory, GENERIC_READ, 0,
 	NULL, OPEN_EXISTING,
@@ -550,7 +624,7 @@ NativeWriteReparse(LinkDirectory, buffer)
     REPARSE_DATA_BUFFER* buffer;
 {
     HANDLE hFile;
-    int returnedLength;
+    DWORD returnedLength;
 
     /* Create the directory - it must not already exist */
     if ((*tclWinProcs->createDirectoryProc)(LinkDirectory, NULL) == 0) {
@@ -568,7 +642,7 @@ NativeWriteReparse(LinkDirectory, buffer)
     }
     /* Set the link */
     if (!DeviceIoControl(hFile, FSCTL_SET_REPARSE_POINT, buffer,
-			 buffer->ReparseDataLength
+			 (DWORD) buffer->ReparseDataLength
 			 + REPARSE_MOUNTPOINT_HEADER_SIZE,
 			 NULL, 0, &returnedLength, NULL)) {
 	/* Error setting junction */
@@ -666,35 +740,48 @@ TclpMatchInDirectory(interp, resultPtr, pathPtr, pattern, types)
 				 * May be NULL. In particular the directory
 				 * flag is very important. */
 {
-    CONST TCHAR *nativeName;
+    CONST TCHAR *native;
 
     if (pattern == NULL || (*pattern == '\0')) {
 	Tcl_Obj *norm = Tcl_FSGetNormalizedPath(NULL, pathPtr);
 	if (norm != NULL) {
+	    /* Match a single file directly */
 	    int len;
-	    char *str = Tcl_GetStringFromObj(norm,&len);
-	    /* Match a file directly */
-	    nativeName = (CONST TCHAR*) Tcl_FSGetNativePath(pathPtr);
-	    if (NativeMatchType(str, len, nativeName, types)) {
+	    DWORD attr;
+	    CONST char *str = Tcl_GetStringFromObj(norm,&len);
+
+	    native = (CONST TCHAR*) Tcl_FSGetNativePath(pathPtr);
+
+	    if (tclWinProcs->getFileAttributesExProc == NULL) {
+		attr = (*tclWinProcs->getFileAttributesProc)(native);
+		if (attr == 0xffffffff) {
+		    return TCL_OK;
+		}
+	    } else {
+		WIN32_FILE_ATTRIBUTE_DATA data;
+		if ((*tclWinProcs->getFileAttributesExProc)(native,
+			GetFileExInfoStandard, &data) != TRUE) {
+		    return TCL_OK;
+		}
+		attr = data.dwFileAttributes;
+	    }
+	    if (NativeMatchType(WinIsDrive(str,len), attr,
+				native, types)) {
 		Tcl_ListObjAppendElement(interp, resultPtr, pathPtr);
 	    }
 	}
 	return TCL_OK;
     } else {
-	char drivePat[] = "?:\\";
-	const char *message;
-	CONST char *dir;
-	char *root;
-	int dirLength;
-	Tcl_DString dirString;
-	DWORD attr, volFlags;
+	DWORD attr;
 	HANDLE handle;
 	WIN32_FIND_DATAT data;
-	BOOL found;
-	Tcl_DString ds;
-	Tcl_DString dsOrig;
-	Tcl_Obj *fileNamePtr;
+	CONST char *dirName;
+	int dirLength;
 	int matchSpecialDots;
+	Tcl_DString ds;        /* native encoding of dir */
+	Tcl_DString dsOrig;    /* utf-8 encoding of dir */
+	Tcl_DString dirString; /* utf-8 encoding of dir with \'s */
+	Tcl_Obj *fileNamePtr;
 
 	/*
 	 * Convert the path to normalized form since some interfaces only
@@ -707,17 +794,16 @@ TclpMatchInDirectory(interp, resultPtr, pathPtr, pattern, types)
 	    return TCL_ERROR;
 	}
 	Tcl_DStringInit(&dsOrig);
-	Tcl_DStringAppend(&dsOrig, Tcl_GetString(fileNamePtr), -1);
+	dirName = Tcl_GetStringFromObj(fileNamePtr, &dirLength);
+	Tcl_DStringAppend(&dsOrig, dirName, dirLength);
 
-	dirLength = Tcl_DStringLength(&dsOrig);
 	Tcl_DStringInit(&dirString);
 	if (dirLength == 0) {
 	    Tcl_DStringAppend(&dirString, ".\\", 2);
 	} else {
 	    char *p;
 
-	    Tcl_DStringAppend(&dirString, Tcl_DStringValue(&dsOrig),
-		    Tcl_DStringLength(&dsOrig));
+	    Tcl_DStringAppend(&dirString, dirName, dirLength);
 	    for (p = Tcl_DStringValue(&dirString); *p != '\0'; p++) {
 		if (*p == '/') {
 		    *p = '\\';
@@ -731,14 +817,16 @@ TclpMatchInDirectory(interp, resultPtr, pathPtr, pattern, types)
 		dirLength++;
 	    }
 	}
-	dir = Tcl_DStringValue(&dirString);
+	dirName = Tcl_DStringValue(&dirString);
+	Tcl_DecrRefCount(fileNamePtr);
 
 	/*
 	 * First verify that the specified path is actually a directory.
 	 */
 
-	nativeName = Tcl_WinUtfToTChar(dir, Tcl_DStringLength(&dirString), &ds);
-	attr = (*tclWinProcs->getFileAttributesProc)(nativeName);
+	native = Tcl_WinUtfToTChar(dirName, Tcl_DStringLength(&dirString),
+		&ds);
+	attr = (*tclWinProcs->getFileAttributesProc)(native);
 	Tcl_DStringFree(&ds);
 
 	if ((attr == 0xffffffff) || ((attr & FILE_ATTRIBUTE_DIRECTORY) == 0)) {
@@ -747,51 +835,24 @@ TclpMatchInDirectory(interp, resultPtr, pathPtr, pattern, types)
 	}
 
 	/*
-	 * Next check the volume information for the directory to see
-	 * whether comparisons should be case sensitive or not.  If the
-	 * root is null, then we use the root of the current directory.
-	 * If the root is just a drive specifier, we use the root
-	 * directory of the given drive.
+	 * We need to check all files in the directory, so append a *.*
+	 * to the path.
 	 */
 
-	switch (Tcl_GetPathType(dir)) {
-	    case TCL_PATH_RELATIVE:
-		found = GetVolumeInformationA(NULL, NULL, 0, NULL, NULL,
-			&volFlags, NULL, 0);
-		break;
-	    case TCL_PATH_VOLUME_RELATIVE:
-		if (dir[0] == '\\') {
-		    root = NULL;
-		} else {
-		    root = drivePat;
-		    *root = dir[0];
-		}
-		found = GetVolumeInformationA(root, NULL, 0, NULL, NULL,
-			&volFlags, NULL, 0);
-		break;
-	    case TCL_PATH_ABSOLUTE:
-		if (dir[1] == ':') {
-		    root = drivePat;
-		    *root = dir[0];
-		    found = GetVolumeInformationA(root, NULL, 0, NULL, NULL,
-			    &volFlags, NULL, 0);
-		} else if (dir[1] == '\\') {
-		    char *p;
+	dirName = Tcl_DStringAppend(&dirString, "*.*", 3);
+	native = Tcl_WinUtfToTChar(dirName, -1, &ds);
+	handle = (*tclWinProcs->findFirstFileProc)(native, &data);
+	Tcl_DStringFree(&ds);
 
-		    p = strchr(dir + 2, '\\');
-		    p = strchr(p + 1, '\\');
-		    p++;
-		    nativeName = Tcl_WinUtfToTChar(dir, p - dir, &ds);
-		    found = (*tclWinProcs->getVolumeInformationProc)(nativeName,
-			    NULL, 0, NULL, NULL, &volFlags, NULL, 0);
-		    Tcl_DStringFree(&ds);
-		}
-		break;
-	}
-
-	if (found == 0) {
-	    message = "couldn't read volume information for \"";
-	    goto error;
+	if (handle == INVALID_HANDLE_VALUE) {
+	    Tcl_DStringFree(&dirString);
+	    TclWinConvertError(GetLastError());
+	    Tcl_ResetResult(interp);
+	    Tcl_AppendResult(interp, "couldn't read directory \"",
+		    Tcl_DStringValue(&dsOrig), "\": ",
+		    Tcl_PosixError(interp), (char *) NULL);
+	    Tcl_DStringFree(&dsOrig);
+	    return TCL_ERROR;
 	}
 
 	/*
@@ -811,44 +872,40 @@ TclpMatchInDirectory(interp, resultPtr, pathPtr, pattern, types)
 	}
 
 	/*
-	 * We need to check all files in the directory, so append a *.*
-	 * to the path.
+	 * Now iterate over all of the files in the directory, starting
+	 * with the first one we found.
 	 */
 
-	dir = Tcl_DStringAppend(&dirString, "*.*", 3);
-	nativeName = Tcl_WinUtfToTChar(dir, -1, &ds);
-	handle = (*tclWinProcs->findFirstFileProc)(nativeName, &data);
-	Tcl_DStringFree(&ds);
-
-	if (handle == INVALID_HANDLE_VALUE) {
-	    message = "couldn't read directory \"";
-	    goto error;
-	}
-
-	/*
-	 * Now iterate over all of the files in the directory.
-	 */
-
-	for (found = 1; found != 0;
-		found = (*tclWinProcs->findNextFileProc)(handle, &data)) {
-	    CONST TCHAR *nativeMatchResult;
-	    CONST char *name, *fname;
+	do {
+	    CONST char *utfname;
+	    int checkDrive = 0;
+	    int isDrive;
+	    DWORD attr;
 
 	    if (tclWinProcs->useWide) {
-		nativeName = (CONST TCHAR *) data.w.cFileName;
+		native = (CONST TCHAR *) data.w.cFileName;
+		attr = data.w.dwFileAttributes;
 	    } else {
-		nativeName = (CONST TCHAR *) data.a.cFileName;
+		native = (CONST TCHAR *) data.a.cFileName;
+		attr = data.a.dwFileAttributes;
 	    }
-	    name = Tcl_WinTCharToUtf(nativeName, -1, &ds);
+
+	    utfname = Tcl_WinTCharToUtf(native, -1, &ds);
 
 	    if (!matchSpecialDots) {
 		/* If it is exactly '.' or '..' then we ignore it */
-		if (name[0] == '.') {
-		    if (name[1] == '\0'
-		      || (name[1] == '.' && name[2] == '\0')) {
-			continue;
-		    }
+		if ((utfname[0] == '.') && (utfname[1] == '\0'
+			|| (utfname[1] == '.' && utfname[2] == '\0'))) {
+		    Tcl_DStringFree(&ds);
+		    continue;
 		}
+	    } else if (utfname[0] == '.' && utfname[1] == '.'
+		    && utfname[2] == '\0') {
+		/*
+		 * Have to check if this is a drive below, so we can
+		 * correctly match 'hidden' and not hidden files.
+		 */
+		checkDrive = 1;
 	    }
 
 	    /*
@@ -863,60 +920,38 @@ TclpMatchInDirectory(interp, resultPtr, pathPtr, pattern, types)
 	     * the system.
 	     */
 
-	    nativeMatchResult = NULL;
+	    if (Tcl_StringCaseMatch(utfname, pattern, 1)) {
+		/*
+		 * If the file matches, then we need to process the remainder
+		 * of the path.
+		 */
 
-	    if (Tcl_StringCaseMatch(name, pattern, 1) != 0) {
-		nativeMatchResult = nativeName;
-	    }
-	    Tcl_DStringFree(&ds);
-
-	    if (nativeMatchResult == NULL) {
-		continue;
+		if (checkDrive) {
+		    CONST char *fullname = Tcl_DStringAppend(&dsOrig, utfname,
+			    Tcl_DStringLength(&ds));
+		    isDrive = WinIsDrive(fullname, Tcl_DStringLength(&dsOrig));
+		    Tcl_DStringSetLength(&dsOrig, dirLength);
+		} else {
+		    isDrive = 0;
+		}
+		if (NativeMatchType(isDrive, attr, native, types)) {
+		    Tcl_ListObjAppendElement(interp, resultPtr,
+			    TclNewFSPathObj(pathPtr, utfname,
+				    Tcl_DStringLength(&ds)));
+		}
 	    }
 
 	    /*
-	     * If the file matches, then we need to process the remainder
-	     * of the path.
+	     * Free ds here to ensure that native is valid above.
 	     */
-
-	    name = Tcl_WinTCharToUtf(nativeMatchResult, -1, &ds);
-	    Tcl_DStringAppend(&dsOrig, name, -1);
 	    Tcl_DStringFree(&ds);
-
-	    fname = Tcl_DStringValue(&dsOrig);
-	    nativeName = Tcl_WinUtfToTChar(fname, Tcl_DStringLength(&dsOrig),
-					   &ds);
-
-	    if (NativeMatchType(fname, Tcl_DStringLength(&dsOrig),
-				nativeName, types)) {
-		Tcl_ListObjAppendElement(interp, resultPtr,
-			Tcl_NewStringObj(fname, Tcl_DStringLength(&dsOrig)));
-	    }
-	    /*
-	     * Free ds here to ensure that nativeName is valid above.
-	     */
-
-	    Tcl_DStringFree(&ds);
-
-	    Tcl_DStringSetLength(&dsOrig, dirLength);
-	}
+	} while ((*tclWinProcs->findNextFileProc)(handle, &data) == TRUE);
 
 	FindClose(handle);
 	Tcl_DStringFree(&dirString);
 	Tcl_DStringFree(&dsOrig);
-
 	return TCL_OK;
-
-        error:
-	Tcl_DStringFree(&dirString);
-	TclWinConvertError(GetLastError());
-	Tcl_ResetResult(interp);
-	Tcl_AppendResult(interp, message, Tcl_DStringValue(&dsOrig), "\": ",
-			 Tcl_PosixError(interp), (char *) NULL);
-			 Tcl_DStringFree(&dsOrig);
-	return TCL_ERROR;
     }
-
 }
 
 /*
@@ -972,16 +1007,30 @@ WinIsDrive(
     return 0;
 }
 
-
 /*
+ *----------------------------------------------------------------------
+ *
+ * NativeMatchType --
+ *
  * This function needs a special case for a path which is a root
  * volume, because for NTFS root volumes, the getFileAttributesProc
  * returns a 'hidden' attribute when it should not.
+ *
+ * We never make any calss to a 'get attributes' routine here,
+ * since we have arranged things so that our caller already knows
+ * such information.
+ *
+ * Results:
+ *  0 = file doesn't match
+ *  1 = file matches
+ *
+ *----------------------------------------------------------------------
  */
 static int
 NativeMatchType(
-    CONST char *name,         /* Name */
-    int nameLen,              /* Length of name */
+    int isDrive,              /* Is this a drive */
+    DWORD attr,               /* We already know the attributes
+                               * for the file */
     CONST TCHAR* nativeName,  /* Native path to check */
     Tcl_GlobTypeData *types)  /* Type description to match against */
 {
@@ -990,26 +1039,18 @@ NativeMatchType(
      * want to retrieve this info if it is absolutely necessary
      * because it is an expensive call.  Unfortunately, to deal
      * with hidden files properly, we must always retrieve it.
-     * There are more modern Win32 APIs available which we should
-     * look into.
      */
 
-    DWORD attr = (*tclWinProcs->getFileAttributesProc)(nativeName);
-    if (attr == 0xffffffff) {
-	/* File doesn't exist */
-	return 0;
-    }
-
     if (types == NULL) {
 	/* If invisible, don't return the file */
-	if (attr & FILE_ATTRIBUTE_HIDDEN && !WinIsDrive(name, nameLen)) {
+	if (attr & FILE_ATTRIBUTE_HIDDEN && !isDrive) {
 	    return 0;
 	}
     } else {
-	if (attr & FILE_ATTRIBUTE_HIDDEN && !WinIsDrive(name, nameLen)) {
+	if (attr & FILE_ATTRIBUTE_HIDDEN && !isDrive) {
 	    /* If invisible */
 	    if ((types->perm == 0) ||
-	      !(types->perm & TCL_GLOB_PERM_HIDDEN)) {
+		    !(types->perm & TCL_GLOB_PERM_HIDDEN)) {
 		return 0;
 	    }
 	} else {
@@ -1024,54 +1065,52 @@ NativeMatchType(
 		((types->perm & TCL_GLOB_PERM_RONLY) &&
 			!(attr & FILE_ATTRIBUTE_READONLY)) ||
 		((types->perm & TCL_GLOB_PERM_R) &&
-			(NativeAccess(nativeName, R_OK) != 0)) ||
+			(0 /* File exists => R_OK on Windows */)) ||
 		((types->perm & TCL_GLOB_PERM_W) &&
-			(NativeAccess(nativeName, W_OK) != 0)) ||
+			(attr & FILE_ATTRIBUTE_READONLY)) ||
 		((types->perm & TCL_GLOB_PERM_X) &&
-			(NativeAccess(nativeName, X_OK) != 0))
+			(!(attr & FILE_ATTRIBUTE_DIRECTORY)
+			 && !NativeIsExec(nativeName)))
 		) {
 		return 0;
 	    }
 	}
-	if (types->type != 0) {
-	    Tcl_StatBuf buf;
+	if ((types->type & TCL_GLOB_TYPE_DIR)
+	    && (attr & FILE_ATTRIBUTE_DIRECTORY)) {
+	    /* Quicker test for directory, which is a common case */
+	    return 1;
+	} else if (types->type != 0) {
+	    unsigned short st_mode;
+	    int isExec = NativeIsExec(nativeName);
+
+	    st_mode = NativeStatMode(attr, 0, isExec);
 
-	    if (NativeStat(nativeName, &buf, 0) != 0) {
-		/*
-		 * Posix error occurred, either the file
-		 * has disappeared, or there is some other
-		 * strange error.  In any case we don't
-		 * return this file.
-		 */
-		return 0;
-	    }
 	    /*
 	     * In order bcdpfls as in 'find -t'
 	     */
 	    if (
 		((types->type & TCL_GLOB_TYPE_BLOCK) &&
-			S_ISBLK(buf.st_mode)) ||
+			S_ISBLK(st_mode)) ||
 		((types->type & TCL_GLOB_TYPE_CHAR) &&
-			S_ISCHR(buf.st_mode)) ||
+			S_ISCHR(st_mode)) ||
 		((types->type & TCL_GLOB_TYPE_DIR) &&
-			S_ISDIR(buf.st_mode)) ||
+			S_ISDIR(st_mode)) ||
 		((types->type & TCL_GLOB_TYPE_PIPE) &&
-			S_ISFIFO(buf.st_mode)) ||
+			S_ISFIFO(st_mode)) ||
 		((types->type & TCL_GLOB_TYPE_FILE) &&
-			S_ISREG(buf.st_mode))
+			S_ISREG(st_mode))
 #ifdef S_ISSOCK
 		|| ((types->type & TCL_GLOB_TYPE_SOCK) &&
-			S_ISSOCK(buf.st_mode))
+			S_ISSOCK(st_mode))
 #endif
 		) {
 		/* Do nothing -- this file is ok */
 	    } else {
 #ifdef S_ISLNK
 		if (types->type & TCL_GLOB_TYPE_LINK) {
-		    if (NativeStat(nativeName, &buf, 1) == 0) {
-			if (S_ISLNK(buf.st_mode)) {
-			    return 1;
-			}
+		    st_mode = NativeStatMode(attr, 1, isExec);
+		    if (S_ISLNK(st_mode)) {
+			return 1;
 		    }
 		}
 #endif
@@ -1207,7 +1246,6 @@ TclpGetUserHome(name, bufferPtr)
     return result;
 }
 
-
 /*
  *---------------------------------------------------------------------------
  *
@@ -1272,39 +1310,67 @@ NativeAccess(
     return 0;
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * NativeIsExec --
+ *
+ *	Determines if a path is executable.  On windows this is
+ *	simply defined by whether the path ends in any of ".exe",
+ *	".com", or ".bat"
+ *
+ * Results:
+ *	1 = executable, 0 = not.
+ *
+ *----------------------------------------------------------------------
+ */
 static int
 NativeIsExec(nativePath)
     CONST TCHAR *nativePath;
 {
-    CONST char *p, *path;
-    Tcl_DString ds;
+    if (tclWinProcs->useWide) {
+	CONST WCHAR *path;
+	int len;
 
-    /*
-     * This is really not efficient.  We should be able to examine
-     * the native path directly without converting to UTF.
-     */
-    Tcl_DStringInit(&ds);
-    path = Tcl_WinTCharToUtf(nativePath, -1, &ds);
+	path = (CONST WCHAR*)nativePath;
+	len = wcslen(path);
 
-    p = strrchr(path, '.');
-    if (p != NULL) {
-	p++;
-	/*
-	 * Note: in the old code, stat considered '.pif' files as
-	 * executable, whereas access did not.
-	 */
-	if ((stricmp(p, "exe") == 0)
-		|| (stricmp(p, "com") == 0)
-		|| (stricmp(p, "bat") == 0)) {
-	    /*
-	     * File that ends with .exe, .com, or .bat is executable.
-	     */
+	if (len < 5) {
+	    return 0;
+	}
 
-	    Tcl_DStringFree(&ds);
+	if (path[len-4] != L'.') {
+	    return 0;
+	}
+
+	if ((memcmp((char*)(path+len-3),L"exe",3*sizeof(WCHAR)) == 0)
+	    || (memcmp((char*)(path+len-3),L"com",3*sizeof(WCHAR)) == 0)
+	    || (memcmp((char*)(path+len-3),L"bat",3*sizeof(WCHAR)) == 0)) {
 	    return 1;
 	}
+    } else {
+	CONST char *p;
+
+	/* We are only looking for pure ascii */
+
+	p = strrchr((CONST char*)nativePath, '.');
+	if (p != NULL) {
+	    p++;
+	    /*
+	     * Note: in the old code, stat considered '.pif' files as
+	     * executable, whereas access did not.
+	     */
+	    if ((stricmp(p, "exe") == 0)
+		    || (stricmp(p, "com") == 0)
+		    || (stricmp(p, "bat") == 0)) {
+		/*
+		 * File that ends with .exe, .com, or .bat is executable.
+		 */
+
+		return 1;
+	    }
+	}
     }
-    Tcl_DStringFree(&ds);
     return 0;
 }
 
@@ -1330,9 +1396,24 @@ TclpObjChdir(pathPtr)
 {
     int result;
     CONST TCHAR *nativePath;
+#ifdef __CYGWIN__
+    extern int cygwin_conv_to_posix_path
+	_ANSI_ARGS_((CONST char *, char *));
+    char posixPath[MAX_PATH+1];
+    CONST char *path;
+    Tcl_DString ds;
+#endif /* __CYGWIN__ */
 
     nativePath = (CONST TCHAR *) Tcl_FSGetNativePath(pathPtr);
+#ifdef __CYGWIN__
+    /* Cygwin chdir only groks POSIX path. */
+    path = Tcl_WinTCharToUtf(nativePath, -1, &ds);
+    cygwin_conv_to_posix_path(path, posixPath);
+    result = (chdir(posixPath) == 0 ? 1 : 0);
+    Tcl_DStringFree(&ds);
+#else /* __CYGWIN__ */
     result = (*tclWinProcs->setCurrentDirectoryProc)(nativePath);
+#endif /* __CYGWIN__ */
 
     if (result == 0) {
 	TclWinConvertError(GetLastError());
@@ -1476,9 +1557,13 @@ TclpObjStat(pathPtr, statPtr)
 
     transPtr = Tcl_FSGetTranslatedPath(NULL, pathPtr);
     if (transPtr == NULL || (strpbrk(Tcl_GetString(transPtr), "?*") != NULL)) {
+	if (transPtr != NULL) {
+	    Tcl_DecrRefCount(transPtr);
+	}
 	Tcl_SetErrno(ENOENT);
 	return -1;
     }
+    Tcl_DecrRefCount(transPtr);
 #endif
 
     /*
@@ -1526,7 +1611,8 @@ NativeStat(nativePath, statPtr, checkLinks)
     WCHAR nativeFullPath[MAX_PATH];
     TCHAR *nativePart;
     CONST char *fullPath;
-    int dev, mode;
+    int dev;
+    unsigned short mode;
 
     if (tclWinProcs->getFileAttributesExProc == NULL) {
         /*
@@ -1675,6 +1761,31 @@ NativeStat(nativePath, statPtr, checkLinks)
 	statPtr->st_ctime = ToCTime(data.ftCreationTime);
     }
 
+    mode = NativeStatMode(attr, checkLinks, NativeIsExec(nativePath));
+
+    statPtr->st_dev	= (dev_t) dev;
+    statPtr->st_ino	= 0;
+    statPtr->st_mode	= mode;
+    statPtr->st_nlink	= 1;
+    statPtr->st_uid	= 0;
+    statPtr->st_gid	= 0;
+    statPtr->st_rdev	= (dev_t) dev;
+    return 0;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * NativeStatMode --
+ *
+ *	Calculate just the 'st_mode' field of a 'stat' structure.
+ *
+ *----------------------------------------------------------------------
+ */
+static unsigned short
+NativeStatMode(DWORD attr, int checkLinks, int isExec)
+{
+    int mode;
     if (checkLinks && (attr & FILE_ATTRIBUTE_REPARSE_POINT)) {
 	/* It is a link */
 	mode = S_IFLNK;
@@ -1682,7 +1793,7 @@ NativeStat(nativePath, statPtr, checkLinks)
 	mode  = (attr & FILE_ATTRIBUTE_DIRECTORY) ? S_IFDIR | S_IEXEC : S_IFREG;
     }
     mode |= (attr & FILE_ATTRIBUTE_READONLY) ? S_IREAD : S_IREAD | S_IWRITE;
-    if (NativeIsExec(nativePath)) {
+    if (isExec) {
 	mode |= S_IEXEC;
     }
 
@@ -1693,15 +1804,7 @@ NativeStat(nativePath, statPtr, checkLinks)
 
     mode |= (mode & 0x0700) >> 3;
     mode |= (mode & 0x0700) >> 6;
-
-    statPtr->st_dev	= (dev_t) dev;
-    statPtr->st_ino	= 0;
-    statPtr->st_mode	= (unsigned short) mode;
-    statPtr->st_nlink	= 1;
-    statPtr->st_uid	= 0;
-    statPtr->st_gid	= 0;
-    statPtr->st_rdev	= (dev_t) dev;
-    return 0;
+    return (unsigned short)mode;
 }
 
 static time_t
@@ -2011,11 +2114,13 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 	 * links are not possible.  Both of these assumptions
 	 * appear to be true of these operating systems.
 	 */
-	Tcl_Obj *temp = NULL;
 	int isDrive = 1;
 	Tcl_DString ds;
 
 	currentPathEndPosition = path + nextCheckpoint;
+        if (*currentPathEndPosition == '/') {
+	    currentPathEndPosition++;
+        }
 	while (1) {
 	    char cur = *currentPathEndPosition;
 	    if ((cur == '/' || cur == 0) && (path != currentPathEndPosition)) {
@@ -2085,6 +2190,9 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 	Tcl_DString ds;
 
 	currentPathEndPosition = path + nextCheckpoint;
+	if (*currentPathEndPosition == '/') {
+	    currentPathEndPosition++;
+	}
 	while (1) {
 	    char cur = *currentPathEndPosition;
 	    if ((cur == '/' || cur == 0) && (path != currentPathEndPosition)) {
@@ -2171,7 +2279,7 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 			Tcl_DStringAppend(&dsNorm,(CONST char*)L"/",
 					  sizeof(WCHAR));
 			Tcl_DStringAppend(&dsNorm,(TCHAR*)nativeName,
-					  wcslen(nativeName)*sizeof(WCHAR));
+					  (int) (wcslen(nativeName)*sizeof(WCHAR)));
 		    }
 		}
 		Tcl_DStringFree(&ds);
@@ -2222,3 +2330,37 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
     Tcl_DStringFree(&dsNorm);
     return nextCheckpoint;
 }
+
+/*
+ *---------------------------------------------------------------------------
+ *
+ * TclpUtime --
+ *
+ *	Set the modification date for a file.
+ *
+ * Results:
+ *	0 on success, -1 on error.
+ *
+ * Side effects:
+ *	None.
+ *
+ *---------------------------------------------------------------------------
+ */
+int
+TclpUtime(pathPtr, tval)
+    Tcl_Obj *pathPtr;      /* File to modify */
+    struct utimbuf *tval;  /* New modification date structure */
+{
+    int res;
+    /*
+     * Windows uses a slightly different structure name and, possibly,
+     * contents, so we have to copy the information over
+     */
+    struct _utimbuf buf;
+
+    buf.actime = tval->actime;
+    buf.modtime = tval->modtime;
+
+    res = (*tclWinProcs->utimeProc)(Tcl_FSGetNativePath(pathPtr),&buf);
+    return res;
+}
